@@ -28,8 +28,8 @@
 #' The latter should be a list of length equal to the number of dimensions in \code{x},
 #' where each entry is an integer vector or \code{NULL} (in which case the entirety of the dimension is used).
 #'
-#' \code{\link{OLD_extract_sparse_array}(x, index)} will return a \linkS4class{SparseArraySeed}
-#' containing the indices of non-zero entries in \code{x}, subsetted to the indices in \code{index}.
+#' \code{\link{extract_sparse_array}(x, index)} will return a \linkS4class{COO_SparseArray}
+#' representing the subset of \code{x} corresponding to the indices in \code{index}.
 #' The latter should be a list of the same structure as described for \code{extract_array}.
 #'
 #' \code{\link{type}(x)} will return a string containing the type of the TileDBArraySeed object \code{x}.
@@ -61,7 +61,7 @@
 #' is_sparse,TileDBArraySeed-method
 #' type,TileDBArraySeed-method
 #' extract_array,TileDBArraySeed-method
-#' OLD_extract_sparse_array,TileDBArraySeed-method
+#' extract_sparse_array,TileDBArraySeed-method
 #' DelayedArray,TileDBArraySeed-method
 #' path,TileDBArraySeed-method
 #' chunkdim,TileDBArraySeed-method
@@ -93,10 +93,14 @@ TileDBArraySeed <- function(x, attr) {
     }
 
     obj <- tiledb_array(x)
-    on.exit(tiledb_array_close(obj))
+    on.exit(tiledb_array_close(obj), add=TRUE, after=FALSE)
 
     s <- schema(obj)
-    d <- dim(domain(s))
+    dims <- dimensions(s)
+    doms <- lapply(dims, domain)
+    o <- vapply(doms, function(x) x[1L], 0L)
+    d <- vapply(doms, function(x) diff(x) + 1L, 0L)
+    e <- vapply(dims, tiledb::tile, 0L)
 
     a <- attrs(s)
     if (missing(attr)) {
@@ -104,42 +108,36 @@ TileDBArraySeed <- function(x, attr) {
     } else if (!attr %in% names(a)) {
         stop("'attr' not in the TileDB attributes")
     }
-    
-    my.type <- tiledb:::tiledb_datatype_R_type(datatype(a[[attr]]))
-    if (! my.type %in% c("logical", "double", "integer", "character")) {
+
+    my.type <- tiledb_datatype_R_type(datatype(a[[attr]]))
+    if (!(my.type %in% c("logical", "double", "integer", "character"))) {
         stop("'attr' refers to an unsupported type")
     }
-    
-    meta <- .get_metadata(x, sparse=is.sparse(s))
-    if (my.type=="integer" && identical(meta$type, "logical")) {
-        my.type <- meta$type
+
+    opened <- tiledb_array_open(obj, "READ")
+    on.exit(tiledb_array_close(opened), add=TRUE, after=FALSE) # do I need to do this if we already close it above? I don't know.
+    rtype <- tiledb_get_metadata(opened, "type")
+    if (my.type=="integer" && identical(rtype, "logical")) {
+        my.type <- rtype
     }
 
-    dimnames <- vector("list", length(d))
-    if (!is.null(meta$dimnames)) {
-        dimnames <- meta$dimnames
-    }
-
-    new("TileDBArraySeed", dim=d, dimnames=dimnames, path=x, 
-        sparse=is.sparse(s), attr=attr, type=my.type, extent=meta$extent)
-}
-
-.get_metadata <- function(path, sparse) {
-    obj <- tiledb_array(path)
-    on.exit(tiledb_array_close(obj), add=TRUE)
-    obj <- tiledb_array_open(obj, "READ")
-
-    type <- tiledb_get_metadata(obj, "type")
-
-    dimnames <- tiledb_get_metadata(obj, "dimnames")
+    dimnames <- tiledb_get_metadata(opened, "dimnames")
     if (!is.null(dimnames)) {
         dimnames <- .unpack_dimnames(dimnames)
+    } else {
+        dimnames <- vector("list", length(d))
     }
 
-    D <- dimensions(schema(obj))
-    extent <- vapply(D, tile, 0L)
-
-    list(type=type, dimnames=dimnames, extent=extent)
+    new("TileDBArraySeed", 
+        dim=d,
+        dimnames=dimnames,
+        path=x, 
+        sparse=is.sparse(s),
+        attr=attr,
+        type=my.type,
+        extent=e,
+        offset=o
+    )
 }
 
 #' @importFrom S4Vectors setValidity2
@@ -147,6 +145,19 @@ setValidity2("TileDBArraySeed", function(object) {
     msg <- .common_checks(object)
 
     d <- dim(object)
+    o <- object@offset
+    if (length(o) != length(d)) {
+        msg <- c(msg, "'offset' must have the same length as 'dim'")
+    }
+
+    e <- object@extent
+    if (length(e) != length(d)) {
+        msg <- c(msg, "'extent' must have the same length as 'dim'")
+    }
+    if (!all(e >= 0L)) {
+        msg <- c(msg, "'extent' must contain non-negative integers")
+    }
+
     dn <- dimnames(object)
     if (length(dn)!=length(d)) {
         msg <- c(msg, "'dimnames' must the same length as 'dim'")
@@ -185,23 +196,102 @@ setMethod("path", "TileDBArraySeed", function(object, ...) {
     object@path
 })
 
+.compact_ranges <- function(selected, delta=diff(selected)) {
+    is.not.contig <- which(delta != 1L)
+    cbind(
+        selected[c(1L, is.not.contig + 1L)],
+        selected[c(is.not.contig, length(selected))]
+    )
+}
+
 #' @export
 setMethod("extract_array", "TileDBArraySeed", function(x, index) {
     fill <- switch(type(x), double=0, integer=0L, logical=FALSE)
     d2 <- .get_block_dims(x, index)
-    output <- array(fill, dim=d2)
 
     # Hack to overcome zero-length indices that cause tiledb to throw.
     if (any(d2==0L)) {
-        return(output)
+        return(array(fill, dim=d2))
     }
 
-    obj <- tiledb_array(path(x), attrs=x@attr, query_type="READ", as.data.frame=TRUE)
+    raw.output.type <- if (x@sparse) "data.frame" else "array"
+    obj <- tiledb_array(x@path, attrs=x@attr, query_type="READ", return_as=raw.output.type)
     on.exit(tiledb_array_close(obj))
 
-    df <- .extract_values(obj, index)
-    output[df$indices] <- as(df$values, type(x))
-    output
+    ndim <- length(index)
+    contiguous <- remapping <- expanders <- vector("list", ndim)
+    any.modified <- FALSE
+
+    for (i in seq_len(ndim)) {
+        cur.index <- index[[i]]
+
+        if (is.null(cur.index)) {
+            curd <- x@dim[i]
+            contiguous[[i]] <- cbind(1L, curd) + x@offset[i] - 1L
+
+        } else {
+            original <- cur.index + x@offset[i] - 1L
+            selected <- original
+
+            # Need to account for unsorted or duplicate indices.
+            modified <- FALSE
+            if (is.unsorted(selected)) {
+                modified <- TRUE
+                selected <- sort(selected)
+            }
+            delta <- diff(selected)
+            is.dup <- delta == 0L
+            if (any(is.dup)) {
+                modified <- TRUE
+                selected <- selected[c(TRUE, !is.dup)]
+                delta <- diff(selected)
+            }
+
+            if (modified) {
+                any.modified <- TRUE
+                expanders[[i]] <- match(original, selected)
+            }
+            if (raw.output.type == "data.frame") {
+                remapping[[i]] <- selected
+            }
+
+            contiguous[[i]] <- .compact_ranges(selected, delta)
+        }
+    }
+
+    selected_ranges(obj) <- contiguous 
+    output <- obj[]
+
+    if (raw.output.type == "data.frame") {
+        indices <- matrix(0L, nrow(output), ndim)
+        quick.dim <- integer(ndim) 
+        for (i in seq_len(ndim)) {
+            curremap <- remapping[[i]]
+            if (!is.null(curremap)) {
+                indices[,i] <- match(output[[i]], curremap)
+                quick.dim[i] <- length(curremap)
+            } else {
+                indices[,i] <- output[[i]] - x@offset[i] + 1L
+                quick.dim[i] <- x@dim[i]
+            }
+        }
+        extracted <- array(fill, dim=quick.dim)
+        extracted[indices] <- as(output[[ndim + 1L]], x@type)
+    } else {
+        extracted <- output[[x@attr]]
+        storage.mode(extracted) <- x@type
+    }
+
+    if (any.modified) {
+        for (i in seq_along(expanders)) {
+            if (is.null(expanders[[i]])) {
+                expanders[[i]] <- substitute()
+            }
+        }
+        expanders$drop <- FALSE
+        extracted <- do.call(`[`, c(list(extracted), expanders))
+    }
+    extracted
 })
 
 .get_block_dims <- function(x, index) {
@@ -215,18 +305,53 @@ setMethod("extract_array", "TileDBArraySeed", function(x, index) {
 }
 
 #' @export
-setMethod("OLD_extract_sparse_array", "TileDBArraySeed", function(x, index) {
+#' @importFrom SparseArray extract_sparse_array COO_SparseArray
+setMethod("extract_sparse_array", "TileDBArraySeed", function(x, index) {
     d2 <- .get_block_dims(x, index)
     if (any(d2==0L)) {
-        fill <- switch(type(x), double=0, integer=0L, logical=FALSE)
-        return(SparseArraySeed(d2, nzindex=matrix(0L, 0, length(index)), nzdata=fill[0]))
+        return(COO_SparseArray(d2, nzdata=vector(type(x))))
     }
 
-    obj <- tiledb_array(path(x), attrs=x@attr, query_type="READ")
+    obj <- tiledb_array(x@path, attrs=x@attr, query_type="READ", return_as="data.frame")
     on.exit(tiledb_array_close(obj))
 
-    df <- .extract_values(obj, index)
-    SparseArraySeed(d2, nzindex=df$indices, nzdata=as(df$values, type(x)))
+    ndim <- length(index)
+    contiguous <- remapping <- vector("list", ndim)
+
+    for (i in seq_len(ndim)) {
+        cur.index <- index[[i]]
+
+        if (is.null(cur.index)) {
+            curd <- x@dim[i]
+            contiguous[[i]] <- cbind(1L, curd) + x@offset[i] - 1L
+
+        } else {
+            selected <- cur.index + x@offset[i] - 1L
+            remapping[[i]] <- selected
+
+            # No need to worry about duplicates here.
+            if (is.unsorted(selected)) {
+                o <- order(selected)
+                selected <- selected[o]
+            }
+
+            contiguous[[i]] <- .compact_ranges(selected)
+        }
+    }
+
+    selected_ranges(obj) <- contiguous
+    extracted <- obj[]
+    indices <- matrix(0L, nrow(extracted), ndim)
+    for (i in seq_len(ndim)) {
+        curremap <- remapping[[i]]
+        if (!is.null(curremap)) {
+            indices[,i] <- match(extracted[[i]], curremap)
+        } else {
+            indices[,i] <- extracted[[i]] - x@offset[i] + 1L
+        }
+    }
+
+    COO_SparseArray(d2, nzcoo=indices, nzdata=as(extracted[,ndim + 1L], x@type))
 })
 
 #' @export
@@ -241,47 +366,3 @@ setMethod("DelayedArray", "TileDBArraySeed",
 
 #' @export
 setMethod("matrixClass", "TileDBArray", function(x) "TileDBMatrix")
-
-.format_indices <- function(index) {
-    ndim <- length(index)
-    contiguous <- remapping <- vector("list", ndim)
-
-    for (i in seq_len(ndim)) {
-        cur.index <- index[[i]]
-
-        if (!is.null(cur.index)) {
-            o <- order(cur.index)
-            runs <- rle(cur.index[o])
-
-            rv <- runs$value
-            remapping[[i]] <- list(
-                ref.value=rv,
-                ref.length=runs$length,
-                requested=o
-            )
-    
-            is.not.contig <- which(diff(rv)!=1L)
-            contiguous[[i]] <- cbind(
-                rv[c(1L, is.not.contig + 1L)],
-                rv[c(is.not.contig, length(rv))]
-            )
-        }
-    }
-
-    list(contiguous=contiguous, remapping=remapping)
-}
-
-.extract_values <- function(obj, indices) {
-    index.info <- .format_indices(indices)
-    selected_ranges(obj) <- index.info$contiguous
-    extracted <- obj[]
-
-    # Resolves a 1-to-many mapping between the extracted and requested indices.
-    ndim <- length(indices)
-    output <- remap_indices(as.list(extracted[seq_len(ndim)]), index.info$remapping)
-
-    list(
-        indices=output$indices,
-        values=rep.int(extracted[[ndim + 1L]], output$expand)
-    )
-}
